import importlib
import json
import logging
from pathlib import Path

import pytest
from aiverify_shap_toolbox.algo import Plugin
from aiverify_test_engine.interfaces.idata import IData
from aiverify_test_engine.interfaces.imodel import IModel
from aiverify_test_engine.plugins.enums.model_type import ModelType
from aiverify_test_engine.plugins.enums.plugin_type import PluginType
from aiverify_test_engine.plugins.metadata.plugin_metadata import PluginMetadata
from aiverify_test_engine.plugins.plugins_manager import PluginManager
from aiverify_test_engine.utils.json_utils import remove_numpy_formats
from aiverify_test_engine.utils.simple_progress import SimpleProgress


def test_discover_plugin():
    PluginManager.discover(str(Path(importlib.util.find_spec("aiverify_test_engine").origin).parent.resolve()))


# Variables for testing
valid_data_path = str("../../../user_defined_files/data/sample_bc_credit_data.sav")
valid_model_path = str("../../../user_defined_files/model/sample_bc_credit_sklearn_linear.LogisticRegression.sav")
valid_ground_truth_path = str("../../../user_defined_files//data/sample_bc_credit_data.sav")

test_string = "data_str"
test_int = 1
test_float = 3.142
test_negative_value = -123
test_list = ["data_str"]
test_dict = {"data_str": "data_str"}
test_tuple = ("data_str", "data_str")
test_none = None


class ObjectTest:
    def __init__(self):
        test_discover_plugin()
        (
            data_instance,
            data_serializer_instance,
            data_error_message,
        ) = PluginManager.get_instance(PluginType.DATA, **{"filename": valid_data_path})

        (
            model_instance,
            model_serializer_instance,
            model_error_message,
        ) = PluginManager.get_instance(PluginType.MODEL, **{"filename": valid_model_path})

        (
            ground_truth_instance,
            ground_truth_serializer_instance,
            data_error_message,
        ) = PluginManager.get_instance(PluginType.DATA, **{"filename": valid_ground_truth_path})

        ground_truth = "default"
        model_type = ModelType.CLASSIFICATION
        input_args = {
            "explain_type": "global",
            "background_path": ("../../../user_defined_files/data/" "sample_bc_credit_data.sav"),
            "background_samples": 25,
            "data_samples": 25,
        }
        expected_exception = RuntimeError
        expected_exception_msg = "The algorithm has failed data validation"
        logger_instance = logging.getLogger("PluginTestLogger")
        logger_instance.setLevel(logging.DEBUG)

        input_args["ground_truth"] = ground_truth
        input_args["model_type"] = model_type
        input_args["logger"] = logger_instance
        input_args["project_base_path"] = Path().absolute()

        self._data_instance_and_serializer = (data_instance, data_serializer_instance)
        self._model_instance_and_serializer = (
            model_instance,
            model_serializer_instance,
        )
        self._ground_truth_instance_and_serializer = (
            ground_truth_instance,
            ground_truth_serializer_instance,
        )
        self._ground_truth = ground_truth
        self._model_type = model_type
        self._input_args = input_args
        self._expected_exception = expected_exception
        self._expected_exception_msg = expected_exception_msg


@pytest.fixture
def get_logger_instance(request):
    log_instance = logging.getLogger("TestLogger")
    log_level = logging.getLevelName(request.param)
    log_instance.setLevel(log_level)

    yield log_instance


@pytest.fixture
def get_data_instance_and_serializer(request):
    test_discover_plugin()
    (
        data_instance,
        data_serializer_instance,
        data_error_message,
    ) = PluginManager.get_instance(PluginType.DATA, **{"filename": request.param})
    yield (data_instance, data_serializer_instance)


@pytest.fixture
def get_data_instance_and_serializer_without_ground_truth(request):
    test_discover_plugin()
    (
        data_instance,
        data_serializer_instance,
        data_error_message,
    ) = PluginManager.get_instance(PluginType.DATA, **{"filename": request.param})
    test_object = ObjectTest()
    data_instance.remove_ground_truth(test_object._ground_truth)
    yield (data_instance, data_serializer_instance)


@pytest.fixture
def get_invalid_data_instance(request):
    test_discover_plugin()
    with pytest.raises(Exception) as excinfo:
        (
            data_instance,
            data_serializer_instance,
            data_error_message,
        ) = PluginManager.get_instance(PluginType.DATA, **{"filename": request.param})
    return excinfo


@pytest.fixture
def get_model_instance_and_serializer(request):
    test_discover_plugin()
    (
        model_instance,
        model_serializer_instance,
        model_error_message,
    ) = PluginManager.get_instance(PluginType.MODEL, **{"filename": request.param})
    yield (model_instance, model_serializer_instance)


@pytest.fixture
def get_invalid_model_instance(request):
    test_discover_plugin()
    with pytest.raises(RuntimeError) as excinfo:
        (
            model_instance,
            model_serializer_instance,
            model_error_message,
        ) = PluginManager.get_instance(PluginType.MODEL, **{"filename": request.param})
    return excinfo


@pytest.fixture
def get_ground_truth_instance_and_serializer(request):
    test_discover_plugin()
    (
        ground_truth_instance,
        ground_truth_serializer_instance,
        data_error_message,
    ) = PluginManager.get_instance(PluginType.DATA, **{"filename": request.param})
    return (ground_truth_instance, ground_truth_serializer_instance)


def test_create_plugin_instance_with_all_valid_input():
    test_object = ObjectTest()
    test_plugin = Plugin(
        test_object._data_instance_and_serializer,
        test_object._model_instance_and_serializer,
        test_object._ground_truth_instance_and_serializer,
        test_object._data_instance_and_serializer[0],
        test_object._model_instance_and_serializer[0],
        **test_object._input_args,
    )

    assert isinstance(test_plugin._data_instance, IData)
    assert isinstance(test_plugin._model_instance, IModel)
    assert isinstance(test_plugin._ground_truth_instance, IData)
    assert isinstance(test_plugin._logger, logging.Logger)
    assert isinstance(test_plugin._progress_inst, SimpleProgress)


@pytest.mark.parametrize(
    "invalid_data_instance_type",
    [test_string, test_int, test_float, test_list, test_dict, test_tuple, test_none],
)
def test_init_plugin_instance_with_invalid_data_instance_type(
    invalid_data_instance_type,
):
    test_object = ObjectTest()
    expected_exception_msg = "The algorithm has failed data validation"
    with pytest.raises(test_object._expected_exception) as excinfo:
        Plugin(
            (invalid_data_instance_type, test_object._data_instance_and_serializer[1]),
            test_object._model_instance_and_serializer,
            test_object._ground_truth_instance_and_serializer,
            test_object._data_instance_and_serializer[0],
            test_object._model_instance_and_serializer[0],
            **test_object._input_args,
        )
    assert excinfo.type == test_object._expected_exception
    assert expected_exception_msg in str(excinfo)


@pytest.mark.parametrize(
    "invalid_model_instance_type",
    [test_string, test_int, test_float, test_list, test_dict, test_tuple, test_none],
)
def test_init_plugin_instance_with_invalid_model_instance_type(
    invalid_model_instance_type,
):
    test_object = ObjectTest()
    expected_exception_msg = "The algorithm has failed model validation"
    with pytest.raises(test_object._expected_exception) as excinfo:
        Plugin(
            test_object._data_instance_and_serializer,
            (
                invalid_model_instance_type,
                test_object._model_instance_and_serializer[1],
            ),
            test_object._ground_truth_instance_and_serializer,
            test_object._data_instance_and_serializer[0],
            test_object._model_instance_and_serializer[0],
            **test_object._input_args,
        )
    assert excinfo.type == test_object._expected_exception
    assert expected_exception_msg in str(excinfo)


@pytest.mark.parametrize(
    "invalid_ground_truth_instance_type",
    [test_string, test_int, test_float, test_list, test_dict, test_tuple, test_none],
)
def test_init_plugin_instance_with_invalid_ground_truth_instance_type(
    invalid_ground_truth_instance_type,
):
    test_object = ObjectTest()
    expected_exception_msg = "The algorithm has failed ground truth data validation"
    with pytest.raises(test_object._expected_exception) as excinfo:
        Plugin(
            test_object._data_instance_and_serializer,
            test_object._model_instance_and_serializer,
            (
                invalid_ground_truth_instance_type,
                test_object._ground_truth_instance_and_serializer[1],
            ),
            test_object._data_instance_and_serializer[0],
            test_object._model_instance_and_serializer[0],
            **test_object._input_args,
        )
    assert excinfo.type == test_object._expected_exception
    assert expected_exception_msg in str(excinfo)


@pytest.mark.parametrize("ground_truth", [test_int, test_float, test_list, test_dict, test_tuple])
def test_init_plugin_instance_with_invalid_ground_truth_type(ground_truth):
    test_object = ObjectTest()
    test_object._input_args["ground_truth"] = ground_truth
    expected_exception_msg = "The algorithm has failed ground truth header validation."
    with pytest.raises(test_object._expected_exception) as excinfo:
        Plugin(
            test_object._data_instance_and_serializer,
            test_object._model_instance_and_serializer,
            test_object._ground_truth_instance_and_serializer,
            test_object._data_instance_and_serializer[0],
            test_object._model_instance_and_serializer[0],
            **test_object._input_args,
        )
    assert excinfo.type == test_object._expected_exception
    assert expected_exception_msg in str(excinfo)


@pytest.mark.parametrize("ground_truth", [test_none])
def test_init_plugin_instance_with_none_ground_truth_type(ground_truth):
    test_object = ObjectTest()
    test_object._input_args["ground_truth"] = ground_truth
    expected_exception_msg = "The algorithm has failed ground truth header validation."
    with pytest.raises(test_object._expected_exception) as excinfo:
        Plugin(
            test_object._data_instance_and_serializer,
            test_object._model_instance_and_serializer,
            test_object._ground_truth_instance_and_serializer,
            test_object._data_instance_and_serializer[0],
            test_object._model_instance_and_serializer[0],
            **test_object._input_args,
        )
    assert excinfo.type == test_object._expected_exception
    assert expected_exception_msg in str(excinfo)


@pytest.mark.parametrize(
    "get_data_instance_and_serializer, get_model_instance_and_serializer, get_ground_truth_instance_and_serializer",
    [(valid_data_path, valid_model_path, valid_ground_truth_path)],
    indirect=[
        "get_data_instance_and_serializer",
        "get_model_instance_and_serializer",
        "get_ground_truth_instance_and_serializer",
    ],
)
def test_init_plugin_instance_with_missing_ground_truth(
    get_data_instance_and_serializer,
    get_model_instance_and_serializer,
    get_ground_truth_instance_and_serializer,
):
    data_instance_and_serializer = get_data_instance_and_serializer
    model_instance_and_serializer = get_model_instance_and_serializer
    ground_truth_instance_and_serializer = get_ground_truth_instance_and_serializer
    model_type = ModelType.CLASSIFICATION
    input_args = {}
    expected_exception = RuntimeError
    expected_exception_msg = "The algorithm has failed ground truth header validation."
    logger_instance = logging.getLogger("PluginTestLogger")
    logger_instance.setLevel(logging.DEBUG)

    input_args["model_type"] = model_type
    input_args["logger"] = logger_instance

    with pytest.raises(expected_exception) as excinfo:
        Plugin(
            data_instance_and_serializer,
            model_instance_and_serializer,
            ground_truth_instance_and_serializer,
            data_instance_and_serializer[0],
            model_instance_and_serializer[0],
            **input_args,
        )
    assert excinfo.type == expected_exception
    assert expected_exception_msg in str(excinfo)


@pytest.mark.parametrize(
    "invalid_model_type",
    [
        test_string,
        test_int,
        test_float,
        test_list,
        test_dict,
        test_tuple,
        test_none,
        None,
    ],
)
def test_setup_plugin_instance_with_invalid_model_type(invalid_model_type):
    test_object = ObjectTest()
    test_object._input_args["model_type"] = invalid_model_type
    expected_exception_msg = "The algorithm has failed validation for model type"
    with pytest.raises(test_object._expected_exception) as excinfo:
        Plugin(
            test_object._data_instance_and_serializer,
            test_object._model_instance_and_serializer,
            test_object._ground_truth_instance_and_serializer,
            test_object._data_instance_and_serializer[0],
            test_object._model_instance_and_serializer[0],
            **test_object._input_args,
        )
    assert excinfo.type == test_object._expected_exception
    assert expected_exception_msg in str(excinfo)


@pytest.mark.parametrize(
    "invalid_logger_type",
    [test_string, test_int, test_float, test_list, test_dict, test_tuple],
)
def test_setup_plugin_instance_with_invalid_logger_type(invalid_logger_type):
    test_object = ObjectTest()
    test_object._input_args["logger"] = invalid_logger_type
    expected_exception_msg = "The algorithm has failed to set up logger. The logger type is invalid"
    with pytest.raises(test_object._expected_exception) as excinfo:
        Plugin(
            test_object._data_instance_and_serializer,
            test_object._model_instance_and_serializer,
            test_object._ground_truth_instance_and_serializer,
            test_object._data_instance_and_serializer[0],
            test_object._model_instance_and_serializer[0],
            **test_object._input_args,
        )
    assert excinfo.type == test_object._expected_exception
    assert expected_exception_msg in str(excinfo)


@pytest.mark.parametrize(
    "log_info",
    [
        ("data_str", "data_str"),
        ("data_str", logging.DEBUG),
        ("data_str", "data_str"),
        (logging.DEBUG, 123),
        (-3.142, -3.142),
        (["data_str"], ["data_str"]),
        ({"data_str": "data_str"}, {"data_str": "data_str"}),
        (("data_str", "data_str")),
        (None, None),
    ],
)
def test_setup_plugin_instance_with_invalid_log_level_and_message(log_info):
    test_object = ObjectTest()
    expected_exception_msg = "The algorithm has invalid log level or message."
    test_plugin = Plugin(
        test_object._data_instance_and_serializer,
        test_object._model_instance_and_serializer,
        test_object._ground_truth_instance_and_serializer,
        test_object._data_instance_and_serializer[0],
        test_object._model_instance_and_serializer[0],
        **test_object._input_args,
    )
    with pytest.raises(test_object._expected_exception) as excinfo:
        test_plugin.add_to_log(log_info[0], log_info[1])
    assert excinfo.type == test_object._expected_exception
    assert expected_exception_msg in str(excinfo)


@pytest.mark.parametrize(
    "invalid_project_base_path",
    [test_string, test_int, test_float, test_list, test_dict, test_tuple],
)
def test_init_plugin_instance_with_invalid_project_base_path(invalid_project_base_path):
    test_object = ObjectTest()
    test_object._input_args["project_base_path"] = invalid_project_base_path
    expected_exception_msg = "The algorithm has failed validation for the project path."
    with pytest.raises(test_object._expected_exception) as excinfo:
        Plugin(
            test_object._data_instance_and_serializer,
            test_object._model_instance_and_serializer,
            test_object._ground_truth_instance_and_serializer,
            test_object._data_instance_and_serializer[0],
            test_object._model_instance_and_serializer[0],
            **test_object._input_args,
        )
    assert excinfo.type == test_object._expected_exception
    assert expected_exception_msg in str(excinfo)


def test_plugin_valid_get_metadata():
    test_object = ObjectTest()
    test_plugin = Plugin(
        test_object._data_instance_and_serializer,
        test_object._model_instance_and_serializer,
        test_object._ground_truth_instance_and_serializer,
        test_object._data_instance_and_serializer[0],
        test_object._model_instance_and_serializer[0],
        **test_object._input_args,
    )
    assert isinstance(test_plugin.get_metadata(), PluginMetadata)


def test_plugin_valid_get_plugin_type():
    test_object = ObjectTest()
    test_plugin = Plugin(
        test_object._data_instance_and_serializer,
        test_object._model_instance_and_serializer,
        test_object._ground_truth_instance_and_serializer,
        test_object._data_instance_and_serializer[0],
        test_object._model_instance_and_serializer[0],
        **test_object._input_args,
    )
    assert isinstance(test_plugin.get_plugin_type(), PluginType)


@pytest.mark.parametrize(
    "get_data_instance_and_serializer_without_ground_truth",
    [(valid_data_path)],
    indirect=["get_data_instance_and_serializer_without_ground_truth"],
)
def test_valid_run(get_data_instance_and_serializer_without_ground_truth):
    test_object = ObjectTest()
    test_object._ground_truth_instance_and_serializer[0].keep_ground_truth(test_object._ground_truth)
    print(f"[dDATA FEATURES]: {get_data_instance_and_serializer_without_ground_truth[0].read_labels()}")
    print(f"[dGROUND_TRUTH FEATURES]: {test_object._ground_truth_instance_and_serializer[0].read_labels()}")

    test_plugin = Plugin(
        get_data_instance_and_serializer_without_ground_truth,
        test_object._model_instance_and_serializer,
        test_object._ground_truth_instance_and_serializer,
        test_object._data_instance_and_serializer[0],
        test_object._model_instance_and_serializer[0],
        **test_object._input_args,
    )
    test_plugin.generate()
    results = remove_numpy_formats(test_plugin.get_results())
    print("my results", results)

    # Load sample JSON file to assert results
    f = open(str(Path(__file__).parent / "sample_output.json"))
    sample_data = json.load(f)
    f.close()
    
    def approx_nested(struct1, struct2, rel=1e-8):
        """
        Recursively compare nested dictionaries and lists and apply pytest.approx to floating-point numbers.
        """
        if isinstance(struct1, dict) and isinstance(struct2, dict):
            # Recursively compare dictionaries
            return all(approx_nested(struct1[key], struct2[key], rel) for key in struct1)
        elif isinstance(struct1, list) and isinstance(struct2, list):
            # Recursively compare lists element by element
            if len(struct1) != len(struct2):
                return False
            return all(approx_nested(item1, item2, rel) for item1, item2 in zip(struct1, struct2))
        elif isinstance(struct1, (float, int)) and isinstance(struct2, (float, int)):
            # Use pytest.approx to compare numbers with relative tolerance
            return struct1 == pytest.approx(struct2, rel=rel)
        else:
            # Direct comparison for non-numeric values
            return struct1 == struct2

    assert approx_nested(results, sample_data)







from pathlib import Path

import pytest
from aiverify_shap_toolbox.algo_init import AlgoInit
from aiverify_test_engine.plugins.enums.model_type import ModelType

binary_non_pipeline = {
    "data_path": str("../../../user_defined_files/data/sample_bc_credit_data.sav"),
    "model_path": str("../../../user_defined_files/model/sample_bc_credit_sklearn_linear.LogisticRegression.sav"),
    "ground_truth_path": str("../../../user_defined_files/data/sample_bc_credit_data.sav"),
    "run_pipeline": False,
    "model_type": ModelType.CLASSIFICATION,
    "ground_truth": "default",
    "background_path": str("../../../user_defined_files/data/sample_bc_credit_data.sav"),
}

multiclass_non_pipeline = {
    "data_path": str("../../../user_defined_files/data/sample_mc_toxic_data.sav"),
    "model_path": str("../../../user_defined_files/model/sample_bc_credit_sklearn_linear.LogisticRegression.sav"),
    "ground_truth_path": str("../../../user_defined_files/data/sample_mc_toxic_data.sav"),
    "run_pipeline": False,
    "model_type": ModelType.CLASSIFICATION,
    "ground_truth": "toxic",
    "background_path": str("../../../user_defined_files/data/sample_mc_toxic_data.sav"),
}

tabular_binary_pipeline = {
    "data_path": str("../../../user_defined_files/data/sample_bc_pipeline_credit_data.sav"),
    "model_path": str("../../../user_defined_files/pipeline/bc_tabular_credit"),
    "ground_truth_path": str("../../../user_defined_files/data/sample_bc_pipeline_credit_ytest_data.sav"),
    "run_pipeline": True,
    "model_type": ModelType.CLASSIFICATION,
    "ground_truth": "default",
    "background_path": str("../../../user_defined_files/data/sample_bc_pipeline_credit_data.sav"),
}

multiclass_pipeline = {
    "data_path": str("../../../user_defined_files/data/sample_mc_pipeline_toxic_data.sav"),
    "model_path": str("../../../user_defined_files/pipeline/mc_tabular_toxic"),
    "ground_truth_path": str("../../../user_defined_files/data/sample_mc_pipeline_toxic_ytest_data.sav"),
    "run_pipeline": True,
    "model_type": ModelType.CLASSIFICATION,
    "ground_truth": "toxic",
    "background_path": str("../../../user_defined_files/data/sample_mc_pipeline_toxic_data.sav"),
}

regression_non_pipeline = {
    "data_path": str("../../../user_defined_files/data/sample_reg_donation_data.sav"),
    "model_path": str("../../../user_defined_files/model/sample_reg_donation_sklearn_linear.LinearRegression.sav"),
    "ground_truth_path": str("../../../user_defined_files/data/sample_reg_donation_data.sav"),
    "run_pipeline": False,
    "model_type": ModelType.REGRESSION,
    "ground_truth": "donation",
    "background_path": str("../../../user_defined_files/data/sample_reg_donation_data.sav"),
}

regression_pipeline = {
    "data_path": str("../../../user_defined_files/data/sample_reg_pipeline_data.sav"),
    "model_path": str("../../../user_defined_files/pipeline/regression_tabular_donation"),
    "ground_truth_path": str("../../../user_defined_files/data/sample_reg_pipeline_ytest_data.sav"),
    "run_pipeline": True,
    "model_type": ModelType.REGRESSION,
    "ground_truth": "donation",
    "background_path": str("../../../user_defined_files/data/sample_reg_donation_data.sav"),
}


@pytest.mark.parametrize(
    "data_set",
    [
        binary_non_pipeline,
        tabular_binary_pipeline,
        # multiclass_non_pipeline # TODO : Not working, revisit later
        multiclass_pipeline,
        regression_non_pipeline,
        regression_pipeline,
    ],
)
def test_pshap_toolbox_plugin(data_set):
    # Create an instance of PluginTest with defined paths and arguments and Run.
    plugin_argument_values = {
        "explain_type": "global",
        "background_path": data_set["background_path"],
        "background_samples": 25,
        "data_samples": 25,
    }

    core_modules_path = ""
    plugin_test = AlgoInit(
        data_set["run_pipeline"],
        core_modules_path,
        data_set["data_path"],
        data_set["model_path"],
        data_set["ground_truth_path"],
        data_set["ground_truth"],
        data_set["model_type"],
        plugin_argument_values,
    )
    plugin_test.run()

    json_file_path = Path.cwd() / "output" / "results.json"
    assert json_file_path.exists()



import argparse

from aiverify_shap_toolbox.algo_init import AlgoInit
from aiverify_test_engine.plugins.enums.model_type import ModelType

parser = argparse.ArgumentParser(description="Run the plugin test with specified parameters.")


def run():
    parse_input_args()
    invoke_shap_toolbox_plugin()


def parse_input_args():
    global parser

    parser.add_argument("--data_path", required=True, help="Path to the data file.")
    parser.add_argument("--model_path", required=True, help="Path to the model file.")
    parser.add_argument("--ground_truth_path", required=True, help="Path to the ground truth data file.")
    parser.add_argument(
        "--background_path",
        required=True,
        help="Path to the background path data file.",
    )
    parser.add_argument(
        "--ground_truth",
        required=True,
        help="The ground truth column name in the data.",
    )
    parser.add_argument(
        "--run_pipeline",
        action=argparse.BooleanOptionalAction,
        help="Whether to run the test as a pipeline (default: False).",
    )
    parser.add_argument(
        "--model_type",
        required=True,
        choices=["CLASSIFICATION", "REGRESSION"],
        help="The type of model (CLASSIFICATION or REGRESSION).",
    )
    parser.add_argument(
        "--core_modules_path",
        default="",
        help="Path to the core modules (default: empty).",
    )
    parser.add_argument(
        "--explain_type",
        required=True,
        choices=["global", "local"],
        help="Type of Explainability (global or local)",
    )
    parser.add_argument(
        "--background_samples",
        default=0,
        type=int,
        help="The integer value of background samples.",
    )
    parser.add_argument(
        "--data_samples",
        default=0,
        type=int,
        help="The integer value of data samples.",
    )


def invoke_shap_toolbox_plugin():
    # =====================================================================================
    # NOTE: Do not modify the code below
    # =====================================================================================
    # Perform Plugin Testing

    # Parse the arguments
    args = parser.parse_args()

    # Determine the value of run_pipeline
    if args.run_pipeline is None:
        run_pipeline = False  # Default to False if not provided
    else:
        run_pipeline = args.run_pipeline

    # Map string argument to ModelType enum
    model_type = ModelType[args.model_type]

    print("*" * 20)
    # Debugging prints
    print(
        f"Running with the following arguments:\n"
        f"Data Path: {args.data_path}\n"
        f"Model Path: {args.model_path}\n"
        f"Ground Truth Path: {args.ground_truth_path}\n"
        f"Ground Truth: {args.ground_truth}\n"
        f"Run Pipeline: {run_pipeline}\n"
        f"Model Type: {model_type}\n"
        f"Core Modules Path: {args.core_modules_path}\n"
        f"Background samples value: {args.background_samples}\n"
        f"Data samples value: {args.data_samples}\n"
        f"explain Type: {args.explain_type}\n"
        f"background Path: {args.background_path}\n"
    )
    print("*" * 20)

    plugin_argument_values = {
        "explain_type": args.explain_type,
        "background_path": args.background_path,
        "background_samples": args.background_samples,
        "data_samples": args.data_samples,
    }

    try:
        # Create an instance of AlgoInit with defined paths and arguments and Run.
        plugin_test = AlgoInit(
            run_pipeline,
            args.core_modules_path,
            args.data_path,
            args.model_path,
            args.ground_truth_path,
            args.ground_truth,
            model_type,
            plugin_argument_values,
        )
        plugin_test.run()

    except Exception as exception:
        print(f"Exception caught while running the plugin test: {str(exception)}")


if __name__ == "__main__":
    run()



import logging
from pathlib import Path, PurePath
from typing import Any, Dict, List, Tuple, Union

import numpy as np
import shap
import xgboost
from aiverify_shap_toolbox.utils.explain_types import ExplainType
from aiverify_test_engine.interfaces.ialgorithm import IAlgorithm
from aiverify_test_engine.interfaces.idata import IData
from aiverify_test_engine.interfaces.imodel import IModel
from aiverify_test_engine.interfaces.ipipeline import IPipeline
from aiverify_test_engine.interfaces.iserializer import ISerializer
from aiverify_test_engine.plugins.enums.data_plugin_type import DataPluginType
from aiverify_test_engine.plugins.enums.model_plugin_type import ModelPluginType
from aiverify_test_engine.plugins.enums.pipeline_plugin_type import PipelinePluginType
from aiverify_test_engine.plugins.enums.model_type import ModelType
from aiverify_test_engine.plugins.enums.plugin_type import PluginType
from aiverify_test_engine.plugins.metadata.plugin_metadata import PluginMetadata
from aiverify_test_engine.plugins.plugins_manager import PluginManager
from aiverify_test_engine.utils.json_utils import load_schema_file, validate_json
from aiverify_test_engine.utils.simple_progress import SimpleProgress
from aiverify_test_engine.utils.url_utils import is_url


# =====================================================================================
# NOTE:
# 1. Check that you have installed the aiverify_test_engine latest package.
# 2. Check that you have run tests/install_core_plugins_requirements.sh to install all the
#    requirements required by the core plugins (serializers, data, models).
#    Alternatively, you may install the plugins that you require by installing the
#    requirements individually.
# 3. Do not modify the class name, else the plugin cannot be read by the system.
# =====================================================================================
class Plugin(IAlgorithm):
    """
    The Plugin(SHAP Toolbox) class specifies methods in generating results for algorithm
    """

    # Some information on plugin
    _name: str = "SHAP Toolbox"
    _description: str = (
        "SHAP (SHapley Additive exPlanations) is a game theoretic approach to explain the "
        "output of any machine learning model. It connects optimal credit allocation with "
        "local explanations using the classic Shapley values from game theory and their related "
        "extensions (see papers for details and citations)."
    )
    _version: str = "0.9.0"
    _metadata: PluginMetadata = PluginMetadata(_name, _description, _version)
    _plugin_type: PluginType = PluginType.ALGORITHM
    _requires_ground_truth: bool = True
    _supported_algorithm_model_type: List = [
        ModelType.CLASSIFICATION,
        ModelType.REGRESSION,
    ]

    @staticmethod
    def get_metadata() -> PluginMetadata:
        """
        A method to return the metadata for this plugin

        Returns:
            PluginMetadata: Metadata of this plugin
        """
        return Plugin._metadata

    @staticmethod
    def get_plugin_type() -> PluginType:
        """
        A method to return the type for this plugin

        Returns:
            PluginType: Type of this plugin
        """
        return Plugin._plugin_type

    def __init__(
        self,
        data_instance_and_serializer: Tuple[IData, ISerializer],
        model_instance_and_serializer: Tuple[IModel, ISerializer],
        ground_truth_instance_and_serializer: Tuple[IData, ISerializer],
        initial_data_instance: Union[IData, None],
        initial_model_instance: Union[IModel, IPipeline, None],
        **kwargs,
    ):
        self._initial_data_instance = initial_data_instance
        self._initial_model_instance = initial_model_instance

        # Look for kwargs values for log_instance, progress_callback and base path
        self._logger = kwargs.get("logger", None)
        self._progress_inst = SimpleProgress(1, 0, kwargs.get("progress_callback", None))

        # Check if data and model are tuples and if the tuples contain 2 items
        if not isinstance(data_instance_and_serializer, Tuple) or len(data_instance_and_serializer) != 2:
            self.add_to_log(
                logging.ERROR,
                f"The algorithm has failed data validation: {data_instance_and_serializer}",
            )
            raise RuntimeError("The algorithm has failed data validation")

        if not isinstance(model_instance_and_serializer, Tuple) or len(model_instance_and_serializer) != 2:
            self.add_to_log(
                logging.ERROR,
                f"The algorithm has failed model validation: {model_instance_and_serializer}",
            )
            raise RuntimeError("The algorithm has failed model validation")

        self._data_instance = data_instance_and_serializer[0]
        self._model_instance = model_instance_and_serializer[0]
        self._model_type = kwargs.get("model_type")

        if Plugin._requires_ground_truth:
            # Check if ground truth instance is tuple and if the tuple contains 2 items
            if (
                not isinstance(ground_truth_instance_and_serializer, Tuple)
                or len(ground_truth_instance_and_serializer) != 2
            ):
                self.add_to_log(
                    logging.ERROR,
                    f"The algorithm has failed ground truth data validation: \
                        {ground_truth_instance_and_serializer}",
                )
                raise RuntimeError("The algorithm has failed ground truth data validation")
            self._requires_ground_truth = True
            self._ground_truth_instance = ground_truth_instance_and_serializer[0]
            self._ground_truth_serializer = ground_truth_instance_and_serializer[1]
            self._ground_truth = kwargs.get("ground_truth")

        else:
            self._ground_truth_instance = None
            self._ground_truth = ""

        self._base_path = kwargs.get("project_base_path", Path().absolute())
        self._sensitive_feature = kwargs.get("sensitive_feature", None)

        # Other variables
        self._background = None
        self._background_path = ""
        self._background_instance = None
        self._data = None
        self._data_labels = list()
        self._explain_type = ""
        self._model = None
        self._sample_seed = 1
        self._results = {"results": [0]}

        # Perform setup for this plug-in
        self.setup()

        # Algorithm input schema defined in input.schema.json
        # By defining the input schema, it allows the front-end to know what algorithm input params is
        # required by this plugin. This allows this algorithm plug-in to receive the arguments values it requires.
        current_file_dir = Path(__file__).parent
        self._input_schema = load_schema_file(str(current_file_dir / "input.schema.json"))

        # Algorithm output schema defined in output.schema.json
        # By defining the output schema, this plug-in validates the result with the output schema.
        # This allows the result to be validated against the schema before passing it to the front-end for display.
        self._output_schema = load_schema_file(str(current_file_dir / "output.schema.json"))

        # Retrieve the input parameters defined in the input schema and store them
        self._input_arguments = dict()
        for key in self._input_schema.get("properties").keys():
            self._input_arguments.update({key: kwargs.get(key)})

        print(f"_input_arguments :  {self._input_arguments}")
        # Perform validation on input argument schema
        if not validate_json(self._input_arguments, self._input_schema):
            self.add_to_log(
                logging.ERROR,
                f"The algorithm has failed input schema validation. \
                    The input must adhere to the schema in input.schema.json: {self._input_arguments}",
            )
            raise RuntimeError(
                "The algorithm has failed input schema validation. \
                The input must adhere to the schema in input.schema.json"
            )

    def add_to_log(self, log_level: int, log_message: str) -> None:
        """
        A helper method to log messages to store events occurred

        Args:
            log_level (int): The logging level
            log_message (str): The logging message
        """
        if self._logger is not None:
            if not isinstance(log_level, int) or not isinstance(log_message, str):
                raise RuntimeError(
                    "The algorithm has invalid log level or message. The log level should be a \
                        logging level(i.e. logging.DEBUG) and log message should be in String format"
                )
        if self._logger is not None:
            if log_level is logging.DEBUG:
                self._logger.debug(log_message)
            elif log_level is logging.INFO:
                self._logger.info(log_message)
            elif log_level is logging.WARNING:
                self._logger.warning(log_message)
            elif log_level is logging.ERROR:
                self._logger.error(log_message)
            elif log_level is logging.CRITICAL:
                self._logger.critical(log_message)
            else:
                pass  # Invalid log level
        else:
            pass  # No log instance

    def setup(self) -> None:
        """
        A method to perform setup for this algorithm plugin
        """
        # Perform validation on logger
        if self._logger:
            if not isinstance(self._logger, logging.Logger):
                raise RuntimeError("The algorithm has failed to set up logger. The logger type is invalid")

        # Perform validation on model type
        if self._model_type not in Plugin._supported_algorithm_model_type:
            self.add_to_log(
                logging.ERROR,
                f"The algorithm has failed validation for model type: {self._model_type}",
            )
            raise RuntimeError("The algorithm has failed validation for model type")

        # Perform validation on data instance
        if not isinstance(self._data_instance, IData):
            self.add_to_log(
                logging.ERROR,
                f"The algorithm has failed data validation: {self._data_instance}",
            )
            raise RuntimeError("The algorithm has failed data validation.")

        # Perform validation on model instance
        if not isinstance(self._model_instance, IModel) and not isinstance(self._model_instance, IPipeline):
            self.add_to_log(
                logging.ERROR,
                f"The algorithm has failed model validation: {self._model_instance}",
            )
            raise RuntimeError("The algorithm has failed model validation")

        # Perform validation on ground truth instance
        if self._requires_ground_truth:
            if not isinstance(self._ground_truth_instance, IData):
                self.add_to_log(
                    logging.ERROR,
                    f"The algorithm has failed ground truth data validation: {self._ground_truth_instance}",
                )
                raise RuntimeError("The algorithm has failed ground truth data validation")

            # Perform validation on ground truth header
            if not isinstance(self._ground_truth, str):
                self.add_to_log(
                    logging.ERROR,
                    "The algorithm has failed ground truth header validation. \
                    Header must be in String and must be present in the dataset: {self._ground_truth}",
                )
                raise RuntimeError(
                    "The algorithm has failed ground truth header validation. \
                    Header must be in String and must be present in the dataset"
                )

        # Perform validation on progress_inst
        if self._progress_inst:
            if not isinstance(self._progress_inst, SimpleProgress):
                raise RuntimeError("The algorithm has failed validation for the progress bar")

        # Perform validation on project_base_path
        if not isinstance(self._base_path, PurePath):
            self.add_to_log(
                logging.ERROR,
                "The algorithm has failed validation for the project path. \
                Ensure that the project path is a valid path: {self._base_path}",
            )
            raise RuntimeError(
                "The algorithm has failed validation for the project path. \
                Ensure that the project path is a valid path"
            )

        # Perform validation on metadata
        if not isinstance(self._metadata, PluginMetadata):
            self.add_to_log(
                logging.ERROR,
                f"The algorithm has failed validation for its metadata: {Plugin._metadata}",
            )
            raise RuntimeError("The algorithm has failed validation for its metadata")

        # Perform validation on plugin type
        if not isinstance(self._plugin_type, PluginType):
            self.add_to_log(
                logging.ERROR,
                "The algorithm has failed validation for its plugin type. \
                Ensure that PluginType is PluginType.ALGORITHM: {Plugin._plugin_type}",
            )
            raise RuntimeError(
                "The algorithm has failed validation for its plugin type. \
                Ensure that PluginType is PluginType.ALGORITHM"
            )

        # Perform logging
        self.add_to_log(logging.INFO, "Setup completed")

    def get_progress(self) -> int:
        """
        A method to return the current progress for this plugin

        Returns:
            int: Completion Progress
        """
        return self._progress_inst.get_progress()

    def get_results(self) -> Dict:
        """
        A method to return generated results for this plugin

        Returns:
            Dict: The results to be returned for display
        """
        return self._results

    def generate(self) -> None:
        """
        A method to generate the algorithm results with the provided data, model, ground truth information.
        """
        # Perform SHAP
        try:
            self._explain_shap()
        except Exception:
            import traceback

            traceback.print_exc()

        # Update progress (For 100% completion)
        self._progress_inst.update(1)

    def _explain_shap(self) -> None:
        """
        A helper method to perform shap explanation
        """
        # Get input arguments
        input_background_path = self._input_arguments.get("background_path", "")
        input_background_samples = self._input_arguments.get("background_samples", 0)
        input_data_samples = self._input_arguments.get("data_samples", 0)
        input_explain_type = self._input_arguments.get("explain_type", "")

        # Perform validation on input arguments
        error_count, error_message = self._perform_input_validation(
            input_background_path,
            input_background_samples,
            input_data_samples,
            input_explain_type,
        )
        if error_count > 0:
            self.add_to_log(
                logging.ERROR,
                f"The algorithm has failed input arguments validation: {error_message}",
            )
            raise RuntimeError(f"The algorithm has failed input arguments validation: {error_message}")
        else:
            self.add_to_log(
                logging.INFO,
                f"Validated plugin input arguments - "
                f"Background Path: {input_background_path},"
                f"Background Samples: {input_background_samples},"
                f"Data Samples: {input_data_samples},"
                f"Explain Type: {input_explain_type}",
            )

            # Set the background data
            self._background_path = input_background_path

            # Set the background and data samples
            self._background_samples = input_background_samples
            self._data_samples = input_data_samples

            # Set the Explain Type
            if input_explain_type.lower() == ExplainType.GLOBAL.name.lower():
                self._explain_type = ExplainType.GLOBAL
            else:
                self._explain_type = ExplainType.LOCAL

        # Get background instance
        (
            self._background_instance,
            background_serializer,
            background_error_message,
        ) = PluginManager.get_instance(PluginType.DATA, **{"filename": self._background_path})
        if not self._background_instance:
            self.add_to_log(
                logging.ERROR,
                f"The algorithm failed to get the background data: {self._background_path} - \
                    {background_error_message}",
            )
            raise RuntimeError(
                f"The algorithm failed to get the background data: {self._background_path} - "
                f"{background_error_message}"
            )
        else:
            self.add_to_log(
                logging.INFO,
                f"Background Instance - {self._background_instance},"
                f"Background Serializer - {background_serializer}",
            )

        self._data = self._data_instance.get_data()
        self._data_labels = list(self._data_instance.read_labels().keys())
        self._model = self._initial_model_instance

        # Retrieve data information
        # Check if background dataset has ground truth first
        if self._ground_truth in self._background_instance.get_data().columns:
            self._background = self._background_instance.get_data().drop(self._ground_truth, axis=1)
        else:
            self._background = self._background_instance.get_data()

        if self._model_instance.get_plugin_type() == PluginType.PIPELINE:
            if self._model_instance.get_pipeline_plugin_type() == PipelinePluginType.SKLEARN:
                transformers = self._model.get_pipeline()[:-1]
                self._background = transformers.transform(self._background)

        # Perform data_sampling and background_sampling
        if self._data_samples > 0:
            num_of_samples = min(len(self._data), self._data_samples)
            self._data = self._data.sample(num_of_samples, random_state=self._sample_seed)

            self._ground_truth_data = self._ground_truth_instance.get_data().sample(
                num_of_samples, random_state=self._sample_seed
            )

        if self._background_samples > 0:
            num_of_samples = min(len(self._background), self._background_samples)
            self._background = self._background.sample(num_of_samples, random_state=self._sample_seed)

        # Get explainer function
        explainer = self._get_explainer()
        self.add_to_log(logging.INFO, f"Explainer Name: {explainer}")

        if self._data_instance.get_data_plugin_type() is DataPluginType.PANDAS:
            # Get single shap value
            single_shap_value = explainer.shap_values(self._data.sample(1, random_state=self._sample_seed))

            # Global or Local?
            if self._explain_type is ExplainType.GLOBAL:
                if (
                    self._model_instance.get_plugin_type() is PluginType.MODEL
                    and self._model_instance.get_model_plugin_type() is ModelPluginType.XGBOOST
                ):
                    data = xgboost.DMatrix(self._data, self._ground_truth_data)
                    shap_values = explainer.shap_values(data)
                else:
                    shap_values = explainer.shap_values(self._data)
                results = {
                    "shap_values": shap_values,
                    "samples": self._data,
                    "single_shap_value": single_shap_value,
                    "explainer": explainer,
                }
            else:
                # Local Explain Type
                results = {
                    "single_shap_value": single_shap_value,
                    "explainer": explainer,
                }

            # Format the output results
            output_results = self._format_result(results)

            # Assign the output results
            self._results = output_results

        else:
            self.add_to_log(
                logging.ERROR,
                f"Unsupported data plugin type: {self._data_instance.get_data_plugin_type()}",
            )
            raise RuntimeError(f"Unsupported data plugin type: {self._data_instance.get_data_plugin_type()}")

    def _get_explainer_predict_helper(self, data: Any) -> Any:
        """
        A self-defined function for explainer to perform prediction proba

        Arg:
            data (Any): The data to be sent for prediction proba

        Returns:
            Any: predicted value
        """
        dict_item_labels = self._data_instance.read_labels().items()
        predicted_results = self._model_instance.predict(data, dict_item_labels)
        if isinstance(predicted_results, list):
            predicted_results = np.array(list(predicted_results), dtype="float32")
        return predicted_results

    def _get_explainer(self) -> Union[shap.TreeExplainer, shap.KernelExplainer]:
        """
        A function to return upload explainer. Upload method will return explainer based on algorithm configuration

        Returns:
            Union[shap.TreeExplainer, shap.KernelExplainer]: An instance of either
            TreeExplainer, or KernelExplainer for upload
        """
        if (
            self._model_instance.get_plugin_type() is PluginType.MODEL
            and self._model_instance.get_model_plugin_type() is ModelPluginType.XGBOOST
        ):
            # Tree Explainer
            self.add_to_log(logging.DEBUG, "Using TreeExplainer")
            explainer = shap.TreeExplainer(self._model_instance.get_model())
        else:
            # Others (KernelExplainer)
            self.add_to_log(logging.DEBUG, "Using KernelExplainer")
            if (
                self._model_instance.get_plugin_type() is PluginType.MODEL
                and self._model_instance.get_model_plugin_type() is ModelPluginType.TENSORFLOW
            ):
                explainer = shap.KernelExplainer(self._model_instance.get_model(), self._background)
            else:
                explainer = shap.KernelExplainer(self._get_explainer_predict_helper, self._background)

        return explainer

    def _is_file(self, argument: str) -> bool:
        """
        A helper function to check if argument is a file

        Args:
            argument (str): path to file

        Returns:
            bool: True if argument is a file
        """
        return Path(argument).is_file()

    def _perform_input_validation(
        self,
        input_background_path: str,
        input_background_samples: int,
        input_data_samples: int,
        input_explain_type: str,
    ) -> Tuple[int, str]:
        """
        A helper method to perform input validation

        Args:
            input_background_path (str): Path to background data
            input_background_samples (str): No. of background samples
            input_data_samples (str): No. of data samples
            input_explain_type (str): String of explain type
        Returns:
            Tuple[int, str]: Returns a tuple consisting of total error count and error messages
        """
        error_count = 0
        error_message = ""

        supported_explain_types = [explain_type.name.lower() for explain_type in ExplainType]

        if input_background_path == "" or (
            not self._is_file(input_background_path) and not is_url(input_background_path)
        ):
            error_count += 1
            error_message += "The background path is invalid;"

        if input_background_samples < 0 or not isinstance(input_background_samples, int):
            error_count += 1
            error_message += "The background samples are invalid;"

        if input_data_samples < 0 or not isinstance(input_data_samples, int):
            error_count += 1
            error_message += "The samples are invalid;"

        if input_explain_type == "" or input_explain_type.lower() not in supported_explain_types:
            error_count += 1
            error_message += "The explain type is invalid;"

        return error_count, error_message

    def _format_result(self, results: dict) -> dict:
        """
        A helper method to format the results to match output schema

        Args:
            results (dict): Results to be formatted in dict

        Returns:
            dict: Formatted results in dict
        """
        output_dict = dict({"feature_names": list(), "results": dict()})

        # Populate results dictionary
        output_results = dict(
            {
                "num_local_classes": 0,
                "local": list(),
                "single_explainer_values": [],
                "single_shap_values": list(),
                "global_shap_values": list(),
                "global_samples": list(),
                "global": list(),
                "num_global_classes": 0,
            }
        )

        # Local Explainability
        # Set the local shap value
        # Populate single shap value in class0 and class1 and so on.
        local_shap_values = list()
        for count in range(len(results["single_shap_value"])):
            tmp_value = results["single_shap_value"][count]

            # Convert the tmp_value into an array of dimension.
            # If it is already a ndarray of dimension 2, it will remain same
            single_shap_value = np.array(tmp_value, ndmin=2)
            local_shap_values.append(single_shap_value.tolist())
        output_results.update({"local": local_shap_values})
        output_results.update({"num_local_classes": len(results["single_shap_value"])})

        # Set single explainer values and single shap values
        if isinstance(results["explainer"], list):
            temp_value_ndarray = np.array(results["explainer"][0], ndmin=1)
        else:
            temp_value_ndarray = np.array(results["explainer"].expected_value, ndmin=1)
        output_results.update({"single_explainer_values": temp_value_ndarray.tolist()})

        single_shap_values_list = list()
        for temp_value in results["single_shap_value"]:
            # Convert the tmp_value into an array of dimension 2.
            # If it is already a ndarray of dimension 2, it will remain same
            temp_value_ndarray = np.array(temp_value, ndmin=2)
            # Remove one dimension from the double array
            single_shap_values_list.append(temp_value_ndarray.tolist()[0])
        output_results.update({"single_shap_values": single_shap_values_list})

        # Global bar and force plot
        if self._explain_type is ExplainType.GLOBAL:
            temp_shap_value_ndarray = np.array(results["shap_values"], ndmin=3)
            temp_samples_value_ndarray = np.array(results["samples"].values, ndmin=2)
            output_results.update(
                {
                    "global_shap_values": temp_shap_value_ndarray.tolist(),
                    "global_samples": temp_samples_value_ndarray.tolist(),
                }
            )

            # Calculate average global shap values
            global_avg_shap_values = list()
            global_shap_value_ndarray = np.array(results["shap_values"], ndmin=3)
            num_of_classes = global_shap_value_ndarray.shape[0]
            num_of_features = global_shap_value_ndarray.shape[2]

            for class_count in range(num_of_classes):
                features_average = list()
                for features_count in range(num_of_features):
                    temp_data_slice = global_shap_value_ndarray[class_count, :, features_count]
                    features_average.append(np.abs(temp_data_slice).mean(0))
                global_avg_shap_values.append(features_average)
            output_results.update({"global": global_avg_shap_values})
            output_results.update({"num_global_classes": num_of_classes})

        # Populate feature names
        output_dict["feature_names"] = self._data.columns.tolist()
        output_dict["results"] = output_results

        return output_dict



"""
Allow aiverify_shap_toolbox to be executable through
`python3 -m aiverify_shap_toolbox`
"""

import sys
from importlib.metadata import version
from pathlib import Path

from aiverify_shap_toolbox.plugin_init import run


def main() -> None:
    """
    Print the version of test engine core
    """
    print("*" * 20)
    print(version_msg())
    print("*" * 20)
    # invoke algorithm
    run()


def version_msg():
    """
    Return the aiverify_shap_toolbox version, location and Python powering it.
    """
    python_version = sys.version
    location = Path(__file__).resolve().parent.parent

    return f"Shap toolbox - {version('aiverify_shap_toolbox')} from {location} \
            (Python {python_version})"


if __name__ == "__main__":
    main()



import copy
import importlib
import json
import logging
import sys
import time
from datetime import datetime
from pathlib import Path
from typing import Dict, Tuple, Union

from aiverify_shap_toolbox.algo import Plugin
from aiverify_test_engine.interfaces.idata import IData
from aiverify_test_engine.interfaces.imodel import IModel
from aiverify_test_engine.interfaces.ipipeline import IPipeline
from aiverify_test_engine.interfaces.iserializer import ISerializer
from aiverify_test_engine.interfaces.itestresult import ITestArguments, ITestResult
from aiverify_test_engine.plugins.enums.model_type import ModelType
from aiverify_test_engine.plugins.enums.pipeline_plugin_type import PipelinePluginType
from aiverify_test_engine.plugins.enums.plugin_type import PluginType
from aiverify_test_engine.plugins.plugins_manager import PluginManager
from aiverify_test_engine.utils.json_utils import (
    load_schema_file,
    remove_numpy_formats,
    validate_json,
    validate_test_result_schema,
)
from aiverify_test_engine.utils.time import time_class_method


# =====================================================================================
# NOTE: Do not modify this file unless you know what you are doing.
# =====================================================================================
class AlgoInit:
    """
    The AlgoInit class specifies methods in supporting testing for the plugin.
    """

    @staticmethod
    def progress_callback(completion_value: int):
        """
        A callback function to print the current progress completion

        Args:
            completion_value (int): Current progress completion
        """
        print(f"[AlgoInit] Progress Update: {completion_value}")

    def __init__(
        self,
        run_as_pipeline: bool,
        core_modules_path: str,
        data_path: str,
        model_path: str,
        ground_truth_path: str,
        ground_truth: str,
        model_type: ModelType,
        input_arguments: dict,
    ):
        # Other variables
        self._base_path: Path = Path().absolute()
        self._requires_ground_truth: bool = True
        self._run_as_pipeline: bool = run_as_pipeline
        self._start_time = None
        self._time_taken = None

        # Store the input arguments as private vars
        if core_modules_path == "":
            core_modules_path = Path(importlib.util.find_spec("aiverify_test_engine").origin).parent
        self._core_modules_path: str = core_modules_path
        self._data_path: str = str(data_path)
        self._model_path: str = str(model_path)
        self._input_arguments: Dict = input_arguments
        self._model_type: ModelType = model_type

        if self._requires_ground_truth:
            self._ground_truth_path: str = str(ground_truth_path)
            self._ground_truth: str = ground_truth
        else:
            self._ground_truth_path: str = ""
            self._ground_truth: str = ""

        # Default for instances
        self._initial_data_instance: Union[None, IData] = None
        self._data_instance: Union[None, IData] = None
        self._data_serializer_instance: Union[None, ISerializer] = None

        self._initial_model_instance: Union[None, IModel, IPipeline] = None
        self._model_instance: Union[None, IModel, IPipeline] = None
        self._model_serializer_instance: Union[None, ISerializer] = None

        self._ground_truth_instance: Union[None, IData] = None
        self._ground_truth_serializer_instance: Union[None, ISerializer] = None

        self._logger_instance: Union[None, logging] = None

    @time_class_method
    def run(self) -> None:
        """
        A function to run the plugin test with the provided arguments.
        """
        try:
            print("=" * 20)
            print("START PLUGIN TEST")
            print("=" * 20)

            # Discover available core plugins
            PluginManager.discover(str(self._core_modules_path))
            print(f"[DETECTED_PLUGINS]: {PluginManager.get_printable_plugins()}")

            # Create logger
            self._logger_instance = logging.getLogger("AlgoInitLogger")
            self._logger_instance.setLevel(logging.DEBUG)
            log_format = logging.Formatter(
                fmt="%(levelname)s %(asctime)s \t %(pathname)s %(funcName)s L%(lineno)s - %(message)s",
                datefmt="%Y-%m-%d %H:%M:%S",
            )
            file_handler = logging.FileHandler(filename="test.log")
            file_handler.setFormatter(log_format)
            file_handler.setLevel(level=logging.DEBUG)
            self._logger_instance.addHandler(file_handler)

            if self._run_as_pipeline:
                # Identify and load data information
                (
                    self._data_instance,
                    self._data_serializer_instance,
                    data_error_message,
                ) = PluginManager.get_instance(PluginType.DATA, **{"filename": self._data_path})

                # Identify and load model information
                (
                    self._model_instance,
                    self._model_serializer_instance,
                    model_error_message,
                ) = PluginManager.get_instance(
                    PluginType.PIPELINE,
                    **{"pipeline_path": self._model_path},
                )

                # Perform a copy of the initial data and model information
                self._initial_data_instance = copy.deepcopy(self._data_instance)
                self._initial_model_instance = copy.deepcopy(self._model_instance)

                if self._model_instance.get_pipeline_plugin_type() == PipelinePluginType.SKLEARN:
                    # Perform data transformation
                    current_dataset = self._data_instance.get_data()
                    current_pipeline = self._model_instance.get_pipeline()
                    data_transformation_stages = current_pipeline[:-1]
                    transformed_dataset = data_transformation_stages.transform(current_dataset)
                    transformed_pipeline = current_pipeline[-1]
                    # Set new transformed pipeline and dataset
                    self._data_instance.set_data(transformed_dataset)
                    self._model_instance.set_pipeline(transformed_pipeline)

            else:
                # Get the data, model, and ground truth instance
                (
                    self._data_instance,
                    self._data_serializer_instance,
                    data_error_message,
                ) = PluginManager.get_instance(PluginType.DATA, **{"filename": self._data_path})
                (
                    self._model_instance,
                    self._model_serializer_instance,
                    model_error_message,
                ) = PluginManager.get_instance(PluginType.MODEL, **{"filename": self._model_path})

            # Print the instances we found from the paths and identified from the core plugins
            print(f"[DATA]: {self._data_instance} - {self._data_serializer_instance} ({data_error_message})")
            print(f"[MODEL]: {self._model_instance} - {self._model_serializer_instance} ({model_error_message})")
            print(f"Requires Ground Truth?: {self._requires_ground_truth}")

            if self._data_instance and self._model_instance:
                # Check if ground_truth is optional
                if self._requires_ground_truth:
                    # Get the ground truth instance
                    (
                        self._ground_truth_instance,
                        self._ground_truth_serializer_instance,
                        _ground_truth_error_message,
                    ) = PluginManager.get_instance(PluginType.DATA, **{"filename": self._ground_truth_path})

                    print(f"[GROUND_TRUTH]: {self._ground_truth_instance}{self._ground_truth_serializer_instance}")
                    print(f"[GROUND_TRUTH]: {self._ground_truth}")
                    print(f"[MODEL_TYPE]: {self._model_type}")
                    print(f"[GROUND_TRUTH FEATURES]: {self._ground_truth_instance.read_labels()}")
                    print(self._ground_truth_instance.get_data())

                    print("Removing ground truth from data and keep only ground truth in ground truth data...")

                    # Leave only the ground truth feature in self._ground_truth_instance and
                    # Remove ground truth feature from the data instance
                    is_ground_truth_instance_success = self._ground_truth_instance.keep_ground_truth(self._ground_truth)
                    self._data_instance.remove_ground_truth(self._ground_truth)
                    if not is_ground_truth_instance_success:
                        raise RuntimeError(
                            "ERROR: Unable to retain only ground truth in ground truth instance. (Check "
                            "if "
                            "ground "
                            "truth feature exists in the data specified in ground truth path file.)"
                        )

                    print(f"[GROUND_TRUTH FEATURES]: {self._ground_truth_instance.read_labels()}")

                else:
                    # Do not require Ground Truth
                    self._ground_truth_instance = None

                # Add additional kwargs parameters
                print("Setting additional parameters...")
                self._input_arguments["ground_truth"] = self._ground_truth
                self._input_arguments["model_type"] = self._model_type
                self._input_arguments["logger"] = self._logger_instance
                self._input_arguments["progress_callback"] = AlgoInit.progress_callback
                self._input_arguments["project_base_path"] = self._base_path

                # Run the plugin with the arguments and instances
                print("Creating an instance of the Plugin...")
                self._start_time = time.time()
                plugin = Plugin(
                    (self._data_instance, self._data_serializer_instance),
                    (self._model_instance, self._model_serializer_instance),
                    (self._ground_truth_instance, self._ground_truth_instance),
                    self._initial_data_instance,
                    self._initial_model_instance,
                    **self._input_arguments,
                )

                # Generate the results using this plugin
                print("Generating the results with the plugin...")
                plugin.generate()

                # Get the task results and convert to json friendly and validate against the output schema
                print("Converting numpy formats if exists...")
                results = remove_numpy_formats(plugin.get_results())

                print("Verifying results with output schema...")
                is_success, error_messages = self._verify_task_results(results)
                self._time_taken = time.time() - self._start_time

                if is_success:
                    # Save the output results
                    output_folder = Path.cwd() / "output"
                    output_folder.mkdir(parents=True, exist_ok=True)

                    json_file_path = output_folder / "results.json"
                    self._generate_output_file(results, json_file_path)
                    print("*" * 20)
                    print(f"check the results here : {json_file_path}")
                    print("*" * 20)
                else:
                    raise RuntimeError(error_messages)
            else:
                raise RuntimeError("ERROR: Unable to get data or model instances")

        except Exception as error:
            # Print the error
            print(f"Exception: {str(error)}")

            # Exit with error
            sys.exit(-1)

    def _generate_output_file(self, results, output_path: Path) -> None:
        """
        Format the output results into the AI Verify Test Result and write to a JSON file
        """
        f = open(str(Path(__file__).parent / "algo.meta.json"))
        meta_file = json.load(f)

        # Prepare test arguments
        test_arguments = ITestArguments(
            groundTruth=self._ground_truth,
            modelType=self._model_type.name,
            testDataset=self._data_path,
            modelFile=self._model_path,
            groundTruthDataset=self._ground_truth_path,
            algorithmArgs={
                "run_pipeline": self._run_as_pipeline,
                "explain_type": self._input_arguments["explain_type"],
                "background_samples": self._input_arguments["background_samples"],
                "data_samples": self._input_arguments["data_samples"],
            },
            mode="upload",
        )

        # Create the output result
        output = ITestResult(
            gid=meta_file["gid"],
            cid=meta_file["cid"],
            version=meta_file.get("version"),
            startTime=datetime.fromtimestamp(self._start_time),
            timeTaken=round(self._time_taken, 4),
            testArguments=test_arguments,
            output=results,
        )

        output_json = output.model_dump_json(exclude_none=True, indent=4)
        if validate_test_result_schema(json.loads(output_json)) is True:
            with open(output_path, "w") as json_file:
                json_file.write(output_json)
        else:
            raise RuntimeError("Failed test result schema validation")

    def _verify_task_results(self, task_result: Dict) -> Tuple[bool, str]:
        """
        A helper method to validate the task results according to the output schema

        Args:
            task_result (Dict): A dictionary of results generated by the algorithm

        Returns:
            Tuple[bool, str]: True if validated, False if validation failed.
        """
        is_success = True
        error_message = ""

        # Check that results type is dict
        if type(task_result) is not dict:
            # Raise error - wrong type
            is_success = False
            error_message = f"Invalid type for results: {type(task_result).__name__}"

        else:
            # Validate the json result with the relevant schema.
            # Check that it meets the required format before sending out to the UI for display
            if not validate_json(
                task_result,
                load_schema_file(str(Path(__file__).parent / "output.schema.json")),
            ):
                is_success = False
                error_message = "Failed schema validation"

        return is_success, error_message
